/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2013 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: sample1.c 427 2013-03-05 08:33:45Z ertl-honda $
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include "sample1.h"
#include "../../../../safeg/t_safeg_syscalls.h"
#include "dualoscom.h"

Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))

void task(intptr_t exinf)
{
	int_t	i;
	static int_t	n = 0;
	int_t		tskno = (int_t) exinf;
	const char	*graph[] = { "|", "  +", "	*" };

	syslog(LOG_NOTICE, "task%d is running (%03d).   %s",
				tskno, ++n, graph[tskno-1]);
	for(i=0; i<10000; i++);
}

void btask(intptr_t exinf)
{
	volatile struct safeg_smc_args ret_args;
	safeg_switch_to_nt(&ret_args);
	syslog(LOG_NOTICE, "BTASK should never return!");
}

void cyclic_handler(intptr_t exinf)
{
	SVC_PERROR(iact_tsk(TASK1));
}

static void *memcpy(void *dest, const void *src, size_t count) {
	char *dst8 = (char *)dest;
	char *src8 = (char *)src;

	while (count--) *dst8++ = *src8++;
	return dest;
}

#define SHMEM_SIZE 64

void
share_mem_test() {
	int ret;
	void *buffer;
	uint32_t size;
	int data;
	int i;

	syslog(LOG_NOTICE, "Share mem test start!.");

	ret = dualoscom_shmem_get(DUALOSCOM_SHMEM_1, &buffer, &size);
	if(ret != 0) {
		syslog(LOG_NOTICE, "Get Buffer Error!! %d",ret);
	}

	syslog(LOG_NOTICE, "buffer 0x%x, Size %d.", buffer, size);

	data = 0;
	while(1) {
		for(i = 0; i < SHMEM_SIZE; i++){
			*((uint32_t *)buffer + i) = data + i;
		}
		syslog(LOG_NOTICE, "Write data %d to %d", data, data + SHMEM_SIZE);
		data++;
		tslp_tsk(4000);
	}
}
void
main_task_tex_routine(TEXPTN texptn, intptr_t exinf){
	act_tsk(MAIN_TASK);
	ext_tsk();
}

void main_task(intptr_t exinf)
{
	ER_UINT	ercd;
	int ret;
	dualoscom_block_id_t block_id;
	void *buffer;
	uint32_t size;
	char msg[] = "hello world";

#ifdef DBG_VERBOSE
	SVC_PERROR(syslog_msk_log(0, LOG_UPTO(LOG_DEBUG)));
#else
	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
#endif
	syslog(LOG_NOTICE, "SafeG-COM test program starts (exinf = %d).", (int_t) exinf);

	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID,
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));
#ifdef ENABLE_TASK
	SVC_PERROR(sta_cyc(CYCHDR1));
#endif

#ifdef DISABLE_TIMER
	SVC_PERROR(dis_int(IRQNO_TIMER));
#endif

	ena_tex();
    
	syslog(LOG_NOTICE, "RTOS dualoscom_init");

	ret = dualoscom_init(true, -1);
	assert(ret == 0);

	syslog(LOG_NOTICE, "Dual-OS communications initialized");


	/* phase 1 */
	syslog(LOG_NOTICE, "RTOS send message in channel 0");

	ret = dualoscom_channel_block_alloc(DUALOSCOM_CHANNEL_0, &block_id);
	assert(ret == 0);

	ret = dualoscom_channel_block_getbuffer(DUALOSCOM_CHANNEL_0, block_id, &buffer, &size);
	assert(ret == 0);
	syslog(LOG_NOTICE, "buffer 0x%x, Size %d.", buffer, size);

	assert(sizeof(msg) < size);
	memcpy(buffer, (void *)msg, sizeof(msg));

	ret = dualoscom_channel_block_enqueue(DUALOSCOM_CHANNEL_0, block_id);
	assert(ret == 0);

	syslog(LOG_NOTICE, "RTOS send event");
	ret = dualoscom_channel_event_send(DUALOSCOM_CHANNEL_0);
	assert(ret == 0);

	syslog(LOG_NOTICE, "RTOS wait for ack");
	ret = dualoscom_channel_event_wait(DUALOSCOM_CHANNEL_0, -1);
	assert(ret == 0);


	/* phase 2 */
	syslog(LOG_NOTICE, "RTOS send params 3 and 5");

	*((uint32_t *)buffer) = 3;
	*((uint32_t *)(buffer + 4)) = 5;

	ret = dualoscom_channel_event_send(DUALOSCOM_CHANNEL_0);
	assert(ret == 0);

	syslog(LOG_NOTICE, "RTOS wait result");
	ret = dualoscom_channel_event_wait(DUALOSCOM_CHANNEL_0, -1);
	assert(ret == 0);

	syslog(LOG_NOTICE, "RTOS rpc result: 3 x 5 = %u", *((uint32_t *)buffer));


	/* phase 3 */
	syslog(LOG_NOTICE, "RTOS send params 4 and 6");

	*((uint32_t *)buffer) = 4;
	*((uint32_t *)(buffer + 4)) = 6;

	ret = dualoscom_channel_event_send(DUALOSCOM_CHANNEL_0);
	assert(ret == 0);

	syslog(LOG_NOTICE, "RTOS wait result");
	ret = dualoscom_channel_event_wait(DUALOSCOM_CHANNEL_0, -1);
	assert(ret == 0);

	syslog(LOG_NOTICE, "RTOS rpc result: 4 x 6 = %u", *((uint32_t *)buffer));


	/* phase 4 */
	syslog(LOG_NOTICE, "RTOS wait for sentence from GPOS in channel 1");
	ret = dualoscom_channel_event_wait(DUALOSCOM_CHANNEL_1, -1);
	assert(ret == 0);

	ret = dualoscom_channel_block_dequeue(DUALOSCOM_CHANNEL_1, &block_id);
	assert(ret == 0);

	ret = dualoscom_channel_block_getbuffer(DUALOSCOM_CHANNEL_1, block_id, &buffer, &size);
	assert(ret == 0);

	syslog(LOG_NOTICE, "MSG: %s", (char *)buffer);

	syslog(LOG_NOTICE, "RTOS dualoscom_block_free");
	ret = dualoscom_channel_block_free(DUALOSCOM_CHANNEL_1, block_id);
	assert(ret == 0);


	/* phase 5 */
	syslog(LOG_NOTICE, "RTOS wait for EVEN messages");
	ret = dualoscom_channel_filter_set(DUALOSCOM_CHANNEL_0, DUALOSCOM_FILTER_ID_EVEN);
	assert(ret == 0);

	ret = dualoscom_channel_event_wait(DUALOSCOM_CHANNEL_0, -1);
	assert(ret == 0);

	while(1) {
		ret = dualoscom_channel_block_dequeue(DUALOSCOM_CHANNEL_0, &block_id);
		if (ret == DUALOSCOM_EMPTY) break;
		assert(ret == 0);

		ret = dualoscom_channel_block_getbuffer(DUALOSCOM_CHANNEL_0, block_id, &buffer, &size);
		assert(ret == 0);

		syslog(LOG_NOTICE, "number: %u", *((uint32_t *)buffer));

		ret = dualoscom_channel_block_free(DUALOSCOM_CHANNEL_0, block_id);
		assert(ret == 0);
	}

	syslog(LOG_NOTICE, "RTOS com done!!");

	share_mem_test();

	SVC_PERROR(slp_tsk());

	SVC_PERROR(ext_ker());
	assert(0);
}
