/*
 *  TOPPERS/SafeG Dual-OS monitor
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Safety Gate Dual-OS monitor
 *
 *  Copyright (C) 2012-2013 by Embedded and Real-Time Systems Laboratory
 *     Graduate School of Information Science, Nagoya University, JAPAN
 *
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 *
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 *
 *  @(#) $Id: configurator.c 428 2013-03-05 08:38:58Z ertl-honda $
 */

#include <stdio.h>   /* printf */
#include <string.h>  /* strncpy */
#include <assert.h>  /* assert */
#include <stdint.h>  /* uint32_t */
#include <stdbool.h> /* true, false, bool */
#include <stdlib.h>  /* malloc */
#include <stdarg.h>  /* va_list */
#include <ctype.h>   /* toupper */

#include "linux_list.h" /* list_add_tail */

#define MAX_CHARS 256
#define MAX_PARAM_SIZE_WORDS 4

/*
 * T_OS or NT_OS
 */
#define NUM_OS 2

enum os_t {T_OS = 0, NT_OS = 1};

static char *config_filename[] = {"t_os-dualoscom_config.h", "nt_os-dualoscom_config.h"};

#define SHREGION_CONTROL_SIZE  (4*3)

/*
 * FILTERS
 */
#define DUALOSCOM_FILTER(FILTER_NAME, filter_function, in_t_os, in_nt_os) \
    assert(add_dualoscom_filter(#FILTER_NAME, #filter_function, (in_t_os), (in_nt_os)) == 0);

struct filter_t {
    uint32_t            id[NUM_OS];
    char                filter_name[MAX_CHARS];
    char                function_name[MAX_CHARS];
    bool                is_valid[NUM_OS];
    struct list_head    filter_list;
};

static struct filter_t the_filter_list;
static uint32_t the_num_of_filters[NUM_OS] = {0, 0};

int add_dualoscom_filter(const char *name, const char *func, bool in_t_os, bool in_nt_os)
{
    struct filter_t *filter;

    filter = (struct filter_t *)malloc(sizeof(struct filter_t));
    assert(filter != NULL);

    strncpy(filter->filter_name, name, MAX_CHARS);
    strncpy(filter->function_name, func, MAX_CHARS);
    filter->is_valid[T_OS] = in_t_os;
    filter->is_valid[NT_OS] = in_nt_os;

    if (filter->is_valid[T_OS]) filter->id[T_OS] = the_num_of_filters[T_OS]++;
    if (filter->is_valid[NT_OS]) filter->id[NT_OS] = the_num_of_filters[NT_OS]++;

    list_add_tail(&filter->filter_list, &the_filter_list.filter_list);

    return 0;
}

/*
 * CHANNELS
 */
#define DUALOSCOM_CHANNEL(CHANNEL_NAME, num_blocks, block_size, t_os_default_filter, nt_os_default_filter) \
    assert(add_dualoscom_channel(#CHANNEL_NAME, (num_blocks), (block_size), #t_os_default_filter, #nt_os_default_filter) == 0);

struct channel_t {
    uint32_t            id;
    char                channel_name[MAX_CHARS];
    uint32_t            num_blocks;
    uint32_t            block_size;
    char                default_filter[NUM_OS][MAX_CHARS];
    uint32_t            memory_size;
    struct list_head    channel_list;
};

static struct channel_t the_channel_list;
static uint32_t the_num_of_channels = 0;

int add_dualoscom_channel(const char *name,
                          uint32_t num_blocks,
                          uint32_t block_size,
                          const char *in_t_os_default_filter,
                          const char *in_nt_os_default_filter)
{
    struct channel_t *channel;

    channel = (struct channel_t *)malloc(sizeof(struct channel_t));
    assert(channel != NULL);

    channel->id         = the_num_of_channels++;
    channel->num_blocks = num_blocks;
    channel->block_size = block_size;
    strncpy(channel->channel_name, name, MAX_CHARS);
    strncpy(channel->default_filter[T_OS], in_t_os_default_filter, MAX_CHARS);
    strncpy(channel->default_filter[NT_OS], in_nt_os_default_filter, MAX_CHARS);

    // see dualoscom-shmem.pdf
    channel->memory_size =   8 // T2NT FIFO write and read cursor
                           + 4 // T2NT Filter ID
                           + 8 // NT2T FIFO write and read cursor
                           + 4 // NT2T Filter ID
                           + 4 * (channel->num_blocks + 1) // NT2T FIFO queue
                           + 4 * (channel->num_blocks + 1) // T2NT FIFO queue
                           + channel->num_blocks * (channel->block_size*4 + 4); // locks per block

    list_add_tail(&channel->channel_list, &the_channel_list.channel_list);

    return 0;
}

/*
 * GROUPS
 */
#define DUALOSCOM_CHANNELS_GROUP(GROUP_NAME, num_channels, ...) \
    assert(add_dualoscom_group(#GROUP_NAME, (num_channels), #__VA_ARGS__) == 0);

struct group_t {
    uint32_t            id;
    char                group_name[MAX_CHARS];
    uint32_t            num_channels;
    char                channels_names[MAX_CHARS];
    struct list_head    group_list;
};

static struct group_t the_group_list;
static uint32_t the_num_of_groups = 0;

int add_dualoscom_group(const char *name, uint32_t num_channels, ...)
{
    struct group_t *group;
    va_list ap;
    char *channel_names;

    group = (struct group_t *)malloc(sizeof(struct group_t));
    assert(group != NULL);

    group->id = the_num_of_groups++;
    strncpy(group->group_name, name, MAX_CHARS);
    group->num_channels = num_channels;

    va_start(ap, num_channels);
    channel_names = va_arg(ap, char *);
    va_end(ap);

    strncpy(group->channels_names, channel_names, MAX_CHARS);

    list_add_tail(&group->group_list, &the_group_list.group_list);

    return 0;
}

/*
 * SHMEMS 
 */ 
#define DUALOSCOM_SHMEM(SHMEM_NAME, size) \
    assert(add_dualoscom_shmem(#SHMEM_NAME, (size)) == 0);

struct shmem_t {
    uint32_t            id;
    char                shmem_name[MAX_CHARS];
    uint32_t            buffer_size;
    uint32_t            memory_size;
    struct list_head    shmem_list;
};

static struct shmem_t the_shmem_list;
static uint32_t the_num_of_shmems = 0;

int add_dualoscom_shmem(const char *name,
                          uint32_t buffer_size)
{
    struct shmem_t *shmem;

    shmem = (struct shmem_t *)malloc(sizeof(struct shmem_t));
    assert(shmem != NULL);

    shmem->id          = the_num_of_shmems++;
    shmem->buffer_size = buffer_size;
    strncpy(shmem->shmem_name, name, MAX_CHARS);

    // see dualoscom-shmem.pdf
    shmem->memory_size = 4 // locks per shmem
                           + shmem->buffer_size;

    list_add_tail(&shmem->shmem_list, &the_shmem_list.shmem_list);

    return 0;
}

/*
 * RPCs
 */
#define T2NT 4141
#define NT2T 5959
#define MAX_PARAMS 32

#define DUALOSCOM_RPC(function, direction, parameters) \
    assert(add_dualoscom_rpc(#function, (direction), (parameters)) == 0);

struct rpc_param_t {
    char name[MAX_CHARS];
    bool is_out;
    char type[MAX_CHARS];
};

struct rpc_t {
    uint32_t            id;
    char                function[MAX_CHARS];
    uint32_t            direction;
    uint32_t            num_params;
    struct rpc_param_t  params[MAX_PARAMS];
    char                rpc_channel_name[MAX_CHARS];
    uint32_t            rpc_block_size;
    struct list_head    rpc_list;
};

static struct rpc_t the_rpc_list;
static uint32_t the_num_of_rpcs = 0;

void uppercase(char *string)
{
    while(*string != '\0' ) {
        *string = toupper((unsigned char)*string);
        string++;
    }
}

int add_dualoscom_rpc(const char *function, uint32_t direction, char parameters[])
{
    struct rpc_t *rpc;
    char *par;
    char *params;

    // Allocate a new RPC
    rpc = (struct rpc_t *)malloc(sizeof(struct rpc_t));
    assert(rpc != NULL);

    // set the basic information
    strncpy(rpc->function, function, MAX_CHARS);
    rpc->id        = the_num_of_rpcs++;
    rpc->direction = direction;

    // parse the rpc parameters
    params = (char *)malloc(sizeof(char)*strlen(parameters));
    assert(params != NULL);
    strncpy(params, parameters, sizeof(char)*strlen(parameters));

    rpc->num_params = 0;
    par = strtok(params, " ,:");
    while(par != NULL) {
        // extract parameter name
        strncpy(rpc->params[rpc->num_params].name, par, MAX_CHARS);
        par = strtok(NULL, " ,:");
        assert(par != NULL);
        // extract parameter direction
        rpc->params[rpc->num_params].is_out = (strncmp("out", par, 3) == 0);
        par = strtok(NULL, " ,:");
        assert(par != NULL);
        // extract parameter type (TODO: support params such "unsigned char")
        strncpy(rpc->params[rpc->num_params].type, par, MAX_CHARS);
        par = strtok(NULL, " ,:");
        rpc->num_params++;
    }

    // add rpc to the list
    list_add_tail(&rpc->rpc_list, &the_rpc_list.rpc_list);

    // prepare the channel name (RPC_FUNCTION)
    snprintf(rpc->rpc_channel_name, MAX_CHARS, "RPC_%s", rpc->function);
    uppercase(rpc->rpc_channel_name);

    // calculate the necessary block size in words (TODO: provide a more accurate way)
    rpc->rpc_block_size = rpc->num_params * MAX_PARAM_SIZE_WORDS;

    // add a channel for the RPC
    add_dualoscom_channel(rpc->rpc_channel_name, /* channel name */
                          1,                     /* num_blocks */
                          rpc->rpc_block_size,   /* block size (words) */
                          "NULL_FILTER",         /* t_os_default_filter */
                          "NULL_FILTER"          /* nt_os_default_filter */
                          );
    // free the memory allocated for params
    free(params);

    return 0;
}

/*
 * CONFIG INIT
 */
void config_initialize(void)
{
    INIT_LIST_HEAD(&the_filter_list.filter_list);
    INIT_LIST_HEAD(&the_channel_list.channel_list);
    INIT_LIST_HEAD(&the_shmem_list.shmem_list);
    INIT_LIST_HEAD(&the_group_list.group_list);
    INIT_LIST_HEAD(&the_rpc_list.rpc_list);
#include CONFIG_FILE
}

/*
 * CONFIG GENERATE
 */
void config_generate(enum os_t os)
{
    struct filter_t *filter;
    struct channel_t *channel;
    struct shmem_t *shmem;
    struct group_t *group;
    struct rpc_t *rpc;
    FILE   *fp;
    uint32_t offset;
    bool associated, match;
    struct rpc_param_t *par;
    int i;
    enum os_t dst_os;

    fp = fopen(config_filename[os], "w");
    assert(fp != NULL);

    fprintf(fp,"#ifndef DUALOSCOM_CONFIG_H\n");
    fprintf(fp,"#define DUALOSCOM_CONFIG_H\n");

    fprintf(fp,"\n#include \"dualoscom_hw.h\"\n");
    fprintf(fp,"#include \"dualoscom.h\"\n");

    /*
     *  For User
     */
    fprintf(fp,"\n/* Number of Channel, Shmem, Filter */\n");
    // set number of channels
    fprintf(fp,"#define DUALOSCOM_NUM_CHANNELS %u\n", the_num_of_channels);
    // set number of shmems
    fprintf(fp,"#define DUALOSCOM_NUM_SHMEMS %u\n", the_num_of_shmems);
    // set number of t2nt filters
    fprintf(fp,"#define DUALOSCOM_NUM_T2NT_FILTERS %u\n", the_num_of_filters[T_OS]);
    // set number of nt2t filters
    fprintf(fp,"#define DUALOSCOM_NUM_NT2T_FILTERS %u\n", the_num_of_filters[NT_OS]);

    fprintf(fp,"\n/* ID */\n");
    // set the ids of the channels
    list_for_each_entry(channel, &the_channel_list.channel_list, channel_list) {
        fprintf(fp,"#define %s %u\n", channel->channel_name, channel->id);
    }

    // set the ids of the shmems
    list_for_each_entry(shmem, &the_shmem_list.shmem_list, shmem_list) {
        fprintf(fp,"#define %s %u\n", shmem->shmem_name, shmem->id);
    }

    // set the filter ids (for the other os)
    if (the_num_of_filters[(os+1)%NUM_OS] > 0) {
        list_for_each_entry(filter, &the_filter_list.filter_list, filter_list) {
            if (!filter->is_valid[(os+1)%NUM_OS]) continue;
            fprintf(fp,"#define %s %u\n", filter->filter_name, filter->id[(os+1)%NUM_OS]);
        }
    }

    /*
     * Channels
     */
    fprintf(fp,"\n/* CHANNELS */\n");

    // set the number of blocks
    fprintf(fp,"#define DUALOSCOM_CHANNEL_NUM_BLOCKS {");
    if (the_num_of_channels > 0) {
        list_for_each_entry(channel, &the_channel_list.channel_list, channel_list) {
            if (channel->id != (the_num_of_channels - 1))
                fprintf(fp,"%u, ", channel->num_blocks);
            else
                fprintf(fp,"%u}\n", channel->num_blocks);
        }
    }
    else {
        fprintf(fp,"}\n");
    }
    
    // set the block sizes
    fprintf(fp,"#define DUALOSCOM_CHANNEL_BLOCK_SIZES {");
    if (the_num_of_channels > 0) {
        list_for_each_entry(channel, &the_channel_list.channel_list, channel_list) {
            if (channel->id != (the_num_of_channels - 1))
                fprintf(fp,"%u, ", channel->block_size);
            else
                fprintf(fp,"%u}\n", channel->block_size);
        }
    }
    else {
        fprintf(fp,"}\n");
    }
    
    // set semaphores
    if (os == T_OS) {
        fprintf(fp,"#define DUALOSCOM_CHANNEL_SEMAPHORES {");
        if (the_num_of_channels > 0) {
            list_for_each_entry(channel, &the_channel_list.channel_list, channel_list) {
                if (channel->id != (the_num_of_channels - 1))
                  fprintf(fp,"DUALOSCOM_SEM_CHANNEL_%u, ", channel->id);
                else
                  fprintf(fp,"DUALOSCOM_SEM_CHANNEL_%u}\n", channel->id);
            }
        }
        else {
            fprintf(fp,"}\n");
        }   
    }
    
    // set default filters        
    dst_os = (os == T_OS ? NT_OS : T_OS);
    if (os == T_OS)
      fprintf(fp,"#define DUALOSCOM_CHANNEL_DEFAULT_NT2T_FILTERS {");
    else
      fprintf(fp,"#define DUALOSCOM_CHANNEL_DEFAULT_T2NT_FILTERS {");
    
    if (the_num_of_channels > 0) {
        list_for_each_entry(channel, &the_channel_list.channel_list, channel_list) {
            // make sure that the default filter exists in the OS
            match = false;
            if(strncmp(channel->default_filter[dst_os], "NULL_FILTER", MAX_CHARS) == 0) {
                match = true;
            } else {
                list_for_each_entry(filter, &the_filter_list.filter_list, filter_list) {
                    if(strncmp(channel->default_filter[dst_os], filter->filter_name, MAX_CHARS) == 0) {
                        assert(filter->is_valid[dst_os]);
                        match = true;
                    }
                }
            }
            assert(match);
            
            if (channel->id != (the_num_of_channels - 1))
              fprintf(fp,"%s, ", channel->default_filter[dst_os]);
            else
              fprintf(fp,"%s}\n",  channel->default_filter[dst_os]);
        }
    }
    else {
        fprintf(fp,"}\n");
    }   


    /*
     * Shmems
     */
    fprintf(fp,"\n/* SHMEMS */\n");

    // set the buffer sizes
    fprintf(fp,"#define DUALOSCOM_SHMEM_BUFFER_SIZES {");
    
    if (the_num_of_shmems > 0) {    
        list_for_each_entry(shmem, &the_shmem_list.shmem_list, shmem_list) {
            if (shmem->id != (the_num_of_shmems - 1))
                fprintf(fp,"%u, ", shmem->buffer_size);
            else
                fprintf(fp,"%u}\n", shmem->buffer_size);
        }
    }
    else{
        fprintf(fp,"}\n");
    }

    /*
     * Filters
     */

    fprintf(fp,"\n/* FILTERS */\n");

    if (the_num_of_filters[os] > 0) {
        // define filter function prototypes if they are valid in the current OS
        list_for_each_entry(filter, &the_filter_list.filter_list, filter_list) {
            if (!filter->is_valid[os]) continue;
            fprintf(fp,"extern %s %s(void *buffer, uint32_t size);\n", (os==T_OS)?"bool_t":"bool",filter->function_name);
        }

        // set the names of the filters
        fprintf(fp,"#define DUALOSCOM_FILTERS {");
        list_for_each_entry(filter, &the_filter_list.filter_list, filter_list) {
            if (!filter->is_valid[os]) continue;
            if (filter->id[os] != (the_num_of_filters[os] - 1))
                fprintf(fp,"%s, ", filter->function_name);
            else
                fprintf(fp,"%s}\n", filter->function_name);
        }
    }

    /*
     * Groups
     */

    if (the_num_of_groups > 0) {
        fprintf(fp,"\n/* GROUPS */\n");

        // set number of groups (if there are 0 groups, it is not defined)
        fprintf(fp,"#define DUALOSCOM_NUM_GROUPS %u\n", the_num_of_groups);

        // set the ids of the groups
        list_for_each_entry(group, &the_group_list.group_list, group_list) {
            fprintf(fp,"#define %s %u\n", group->group_name, group->id);
        }
        // set which channels are associated (true) and which are not (false)
        fprintf(fp,"#define DUALOSCOM_CHANNEL2GROUPS {");
        list_for_each_entry(channel, &the_channel_list.channel_list, channel_list) {
            fprintf(fp,"{");
            list_for_each_entry(group, &the_group_list.group_list, group_list) {
                associated = (strstr(group->channels_names, channel->channel_name) != NULL);
                if (group->id != (the_num_of_groups - 1))
                    fprintf(fp,"%s, ", associated?"true":"false");
                else
                    fprintf(fp,"%s}", associated?"true":"false");
            }
            if (channel->id != (the_num_of_channels - 1))
                fprintf(fp,", ");
            else
                fprintf(fp,"}\n");
        }

        // set semaphores
        if (os == T_OS) {
            fprintf(fp,"#define DUALOSCOM_GROUP_SEMAPHORES {");
            list_for_each_entry(group, &the_group_list.group_list, group_list) {
                if (group->id != (the_num_of_groups - 1))
                    fprintf(fp,"DUALOSCOM_SEM_GROUP_%u, ", group->id);
                else
                    fprintf(fp,"DUALOSCOM_SEM_GROUP_%u}\n", group->id);
            }
        }
    }

    /*
     * RPCs
     */

    if (the_num_of_rpcs > 0) {
        fprintf(fp,"\n/* RPCs */\n");

        list_for_each_entry(rpc, &the_rpc_list.rpc_list, rpc_list) {
            // define type structure
            fprintf(fp,"typedef struct {\n");
            for(i=0; i<rpc->num_params; i++) {
                par = &rpc->params[i];
                fprintf(fp,"    %s %s;\n", par->type, par->name);
            }
            fprintf(fp,"} dualoscom_%s_params_t;\n\n", rpc->function);

            // define the rpc stubs
            if (((os == T_OS) && (rpc->direction == T2NT)) ||
                ((os == NT_OS) && (rpc->direction == NT2T))) {
                // client stub
                fprintf(fp,"static inline void %s(", rpc->function);
                for(i=0; i<rpc->num_params; i++) {
                    par = &rpc->params[i];
                    if (i == rpc->num_params-1) {
                        fprintf(fp,"%s %s %s)\n{\n",
                                par->type, par->is_out?"*":"", par->name);
                    } else {
                        fprintf(fp,"%s %s %s, ",
                                par->type, par->is_out?"*":"", par->name);
                    }
                }
                fprintf(fp,"    dualoscom_block_id_t block_id;\n");
                fprintf(fp,"    uint32_t size;\n");
                fprintf(fp,"    dualoscom_%s_params_t *params;\n\n", rpc->function);

                fprintf(fp,"    dualoscom_block_alloc(%s, &block_id);\n", rpc->rpc_channel_name);
                fprintf(fp,"    dualoscom_block_getbuffer(block_id, (void **)&params, &size);\n");
                fprintf(fp,"    assert(sizeof(dualoscom_%s_params_t) < (size*4));\n", rpc->function);

                for(i=0; i<rpc->num_params; i++) {
                    par = &rpc->params[i];
                    if (par->is_out) continue;
                    fprintf(fp,"    params->%s = %s;\n", par->name, par->name);
                }

                fprintf(fp,"    dualoscom_block_enqueue(block_id);\n");
                fprintf(fp,"    dualoscom_channel_event_send(%s);\n", rpc->rpc_channel_name);
                fprintf(fp,"    dualoscom_channel_event_wait(%s, 0);\n", rpc->rpc_channel_name);

                for(i=0; i<rpc->num_params; i++) {
                    par = &rpc->params[i];
                    if (!par->is_out) continue;
                    fprintf(fp,"    *%s = params->%s;\n", par->name, par->name);
                }

                fprintf(fp,"    dualoscom_block_free(block_id);\n}\n");
            } else {
                // server stub
                fprintf(fp,"extern void %s(", rpc->function);
                for(i=0; i<rpc->num_params; i++) {
                    par = &rpc->params[i];
                    if (i == rpc->num_params-1) {
                        fprintf(fp,"%s %s %s);\n\n",
                                par->type, par->is_out?"*":"", par->name);
                    } else {
                        fprintf(fp,"%s %s %s, ",
                                par->type, par->is_out?"*":"", par->name);
                    }
                }

                fprintf(fp,"static inline void server_%s(void)\n{\n", rpc->function);

                fprintf(fp,"    dualoscom_block_id_t block_id;\n");
                fprintf(fp,"    uint32_t size;\n");
                fprintf(fp,"    dualoscom_%s_params_t *params;\n\n", rpc->function);

                fprintf(fp,"    dualoscom_channel_event_wait(%s, 0);\n", rpc->rpc_channel_name);
                fprintf(fp,"    dualoscom_block_dequeue(%s, &block_id);\n", rpc->rpc_channel_name);
                fprintf(fp,"    dualoscom_block_getbuffer(block_id, (void **)&params, &size);\n");
                fprintf(fp,"    assert(sizeof(dualoscom_%s_params_t) < (size*4));\n", rpc->function);

                fprintf(fp,"    %s(", rpc->function);
                for(i=0; i<rpc->num_params; i++) {
                    par = &rpc->params[i];
                    if (i == rpc->num_params-1) {
                        fprintf(fp,"%sparams->%s);\n", par->is_out?"&":"", par->name);
                    } else {
                        fprintf(fp,"%sparams->%s, ", par->is_out?"&":"", par->name);
                    }
                }

                fprintf(fp,"    dualoscom_channel_event_send(%s);\n}\n", rpc->rpc_channel_name);
            }
        }
    }

    /* Shared Regin */
    fprintf(fp,"\n/* SHARED REGION */\n");
    offset = SHREGION_CONTROL_SIZE;
    fprintf(fp,"#define DUALOSCOM_SHREGION_CHANNEL_OFFSETS {");
    if (the_num_of_channels > 0) {
        list_for_each_entry(channel, &the_channel_list.channel_list, channel_list) {
            if (channel->id != (the_num_of_channels - 1))
                fprintf(fp,"%u, ", offset);
            else
                fprintf(fp,"%u", offset);
            offset = offset + channel->memory_size;
        }
    }
    fprintf(fp,"}\n");

    fprintf(fp,"#define DUALOSCOM_SHREGION_PENDING_EVENTS %u\n", offset);
    offset += 8 * the_num_of_channels;
    fprintf(fp,"#define DUALOSCOM_SHREGION_SHMEM_OFFSETS {");
    if (the_num_of_shmems > 0) {
        list_for_each_entry(shmem, &the_shmem_list.shmem_list, shmem_list) {
            if (shmem->id != (the_num_of_shmems - 1))
                fprintf(fp,"%u, ", offset);
            else
                fprintf(fp,"%u", offset);
            offset = offset + shmem->memory_size;
        }
    }
    fprintf(fp,"}\n");

    fprintf(fp,"#define DUALOSCOM_SHREGION_SIZE %u\n", offset);

    fprintf(fp,"\n#endif /* DUALOSCOM_CONFIG_H */\n");
    fclose(fp);
}

void config_asp_cfg(void)
{
    FILE   *fp;
    struct channel_t *channel;
    struct group_t *group;

    /*
     * Create ASP cfg file with semaphores etc.
     */
    fp = fopen("asp-dualoscom.cfg", "w");
    assert(fp != NULL);
    
    fprintf(fp,"#include \"t_os-dualoscom_config.h\"\n\n");

    fprintf(fp,"DEF_INH(NT_INTNO_SOFTIRQ, { TA_NULL, dualoscom_handler });\n");
    fprintf(fp,"CFG_INT(NT_INTNO_SOFTIRQ, { 0, NT_INT_PRIORITY });\n\n");
    
    fprintf(fp,"CRE_SEM(DUALOSCOM_SEM_INIT, { TA_NULL, 0, 1});\n");
    if (the_num_of_channels > 0) {
        list_for_each_entry(channel, &the_channel_list.channel_list, channel_list) {
            fprintf(fp,"CRE_SEM(DUALOSCOM_SEM_CHANNEL_%u, { TA_NULL, 0, 1});\n", channel->id);
        }
    }
    
    if (the_num_of_groups > 0) {
        list_for_each_entry(group, &the_group_list.group_list, group_list) {
            fprintf(fp,"CRE_SEM(DUALOSCOM_SEM_GROUP_%u, { TA_NULL, 0, 1});\n", group->id);
        }
    }
    fclose(fp);
}

/*
 * MAIN
 */
int main()
{
    config_initialize();
    config_generate(T_OS);
    config_generate(NT_OS);
    config_asp_cfg();
    return 0;
}
