/*
 *  TOPPERS/SafeG Dual-OS monitor
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Safety Gate Dual-OS monitor
 *
 *  Copyright (C) 2012-2013 by Embedded and Real-Time Systems Laboratory
 *     Graduate School of Information Science, Nagoya University, JAPAN
 *
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 *
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 *
 *  @(#) $Id: dualoscom.h 428 2013-03-05 08:38:58Z ertl-honda $
 */

#ifndef DUALOSCOM_H
#define DUALOSCOM_H

#include "dualoscom_target.h"

/*
 * ERRORS
 */
#define DUALOSCOM_SUCCESS   (0) // success
#define DUALOSCOM_NOPERM   (-1) // not enough permissions.
#define DUALOSCOM_NOINIT   (-2) // dualos communications system not initialized yet.
#define DUALOSCOM_PARAM    (-3) // incorrect parameter.
#define DUALOSCOM_FULL     (-4) // there are no free blocks.
#define DUALOSCOM_ENQ      (-5) // the block is enqueued.
#define DUALOSCOM_FILTER   (-6) // the block was discarded
#define DUALOSCOM_EMPTY    (-7) // no block is enqueued.
#define DUALOSCOM_ALLOC    (-8) // the block is not allocated.
#define DUALOSCOM_TIMEOUT  (-9) // a timeout occurred
#define DUALOSCOM_LOCKED   (-10) // the shmem is locked
#define DUALOSCOM_UNLOCKED (-11) // the shmem is unlocked
#define DUALOSCOM_ESHREGION (-12) // the shared regison initialize error
#define DUALOSCOM_RESTART   (-13) // restart dualos communications system
#define DUALOSCOM_WAITOVER  (-14) // an other task is waitting for the channel
#define DUALOSCOM_ECTX      (-15) // context error

/*
 * TYPES
 */
typedef uint32_t dualoscom_filter_id_t;
typedef uint32_t dualoscom_channel_id_t;
typedef uint32_t dualoscom_shmem_id_t;
typedef uint32_t dualoscom_block_id_t;
typedef uint32_t dualoscom_group_id_t;
typedef int32_t  dualoscom_time_t;
typedef bool_t (*dualoscom_filter_t)(void *buffer, uint32_t size);

/*
 * Macro
 */
#define NULL_FILTER 0xFFFFFFFF

/*
 *  Channel type
 */
#define CTYPE_CHANNEL 0
#define CTYPE_SHMEM 1


/*
 * dualoscom_init
 *
 * Initializates the Dual-OS Communications system.
 *
 * Errors: DUALOSCOM_PARAM, DUALOSCOM_TIMEOUT, DUALOSCOM_ESHREGION
 *
 */
extern int dualoscom_init(bool_t master, dualoscom_time_t timeout);

/*
 *  API for SHMEM
 */

/*
 * dualoscom_shmem_get
 *
 * To obtain a pointer to the beginning of the memory region of a shmem.
 * 
 * Errors: DUALOSCOM_NOPERM, DUALOSCOM_NOINIT, DUALOSCOM_PARAM
 *
 */
extern int dualoscom_shmem_get(const dualoscom_shmem_id_t shmem_id,
                               void **buffer, uint32_t *size);

/*
 * dualoscom_shmem_trylock
 *
 * Try to get lock of a shmem.
 * 
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM, DUALOSCOM_LOCKED
 *
 */
extern int dualoscom_shmem_trylock(const dualoscom_shmem_id_t shmem_id);

/*
 * dualoscom_shmem_unlock
 *
 * Release lock of a shmem.
 * 
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM, DUALOSCOM_LOCKED
 *
 */
extern int dualoscom_shmem_unlock(const dualoscom_shmem_id_t shmem_id);

/*
 *  API for CHANNEL
 */

/*
 * dualoscom_channel_block_alloc
 *
 * Allocates a block from a channel block pool. This function never blocks the
 * calling task or process.
 *
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM, DUALOSCOM_FULL
 *
 */
extern int dualoscom_channel_block_alloc(const dualoscom_channel_id_t channel_id,
                                         dualoscom_block_id_t *block_id);

/*
 * dualoscom_channel_block_free
 *
 * Releases a block back to the channel block pool where it belongs.
 *
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM, DUALOSCOM_ALLOC
 *
 */
extern int dualoscom_channel_block_free(const dualoscom_channel_id_t channel_id,
                                        const dualoscom_block_id_t   block_id);

/*
 * dualoscom_channel_block_getbuffer
 *
 * To obtain a pointer to the beginning of the memory region of a block.
 *
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM, DUALOSCOM_ALLOC
 *
 */
extern int dualoscom_channel_block_getbuffer(const dualoscom_channel_id_t channel_id,
                                             const dualoscom_block_id_t   block_id,
                                             void                       **buffer,
                                             uint32_t                    *size);

/*
 * dualoscom_channel_block_enqueue
 *
 * Enqueues a block to a channel FIFO. Note that the channel is implicit in
 * the block identifier.
 *
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM, DUALOSCOM_FILTER, DUALOSCOM_ALLOC.
 *
 */
extern int dualoscom_channel_block_enqueue(const dualoscom_channel_id_t channel_id,
                                           const dualoscom_block_id_t   block_id);

/*
 * dualoscom_channel_block_dequeue
 *
 * Dequeues a block from a channel FIFO. This function never blocks the
 * calling task or process.
 *
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM, DUALOSCOM_EMPTY, DUALOSCOM_ALLOC
 *
 */
extern int dualoscom_channel_block_dequeue(const dualoscom_channel_id_t   channel_id,
                                           dualoscom_block_id_t          *block_id);

/*
 * dualoscom_channel_event_send
 *
 * Sends an asynchronous channel event notification. If a notification
 * was already sent and has not been acknowledged yet by the receiver, the
 * function will return as if the notification was sent successfully.
 *
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM
 *
 */
extern int dualoscom_channel_event_send(const dualoscom_channel_id_t channel_id);

/*
 * dualoscom_channel_event_wait
 *
 * This function makes the calling task - or process - wait for an
 * asynchronous event notification on a specific channel. If an event was
 * pending, the function acknowledges it and returns immediately.
 * Otherwise, the calling thread is put in waiting state until an event
 * arrives or a timeout occurs. The units in which the timeout is
 * expressed are implementation-dependent.
 *
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM, DUALOSCOM_TIMEOUT, DUALOSCOM_RESTART
 *
 */
extern int dualoscom_channel_event_wait(const dualoscom_channel_id_t channel_id,
                                        const dualoscom_time_t       timeout);

/*
 * dualoscom_channel_filter_set
 *
 * Used by the receiver to set a communications filter.
 *
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM
 *
 */
extern int dualoscom_channel_filter_set(const dualoscom_channel_id_t channel_id,
                                        const dualoscom_filter_id_t  filter_id);

/*
 * dualoscom_channel_group_event_wait (under development)
 *
 * This function makes the calling task - or process - wait for an
 * asynchronous event notification on a specific group. If an event was
 * pending, the function acknowledges it and returns immediately.
 * Otherwise, the calling thread is put in waiting state until an event
 * arrives or a timeout occurs. The units in which the timeout is
 * expressed are implementation-dependent.
 *
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM, DUALOSCOM_TIMEOUT
 *
 */
extern int dualoscom_channel_group_event_wait(const dualoscom_group_id_t group_id,
                                              const dualoscom_time_t     timeout);

#endif /* DUALOSCOM_H */
