/*
 *  TOPPERS/SafeG Dual-OS monitor
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Safety Gate Dual-OS monitor
 *
 *  Copyright (C) 2012-2013 by Embedded and Real-Time Systems Laboratory
 *     Graduate School of Information Science, Nagoya University, JAPAN
 *
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 *
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 *
 *  @(#) $Id: dualoscom_user.c 428 2013-03-05 08:38:58Z ertl-honda $
 */

#include "dualoscom.h"
#include "dualoscom_kernel.h"

#if (DUALOSCOM_NUM_CHANNELS == 0) && (DUALOSCOM_NUM_SHMEMS == 0)
#error "No channels defined. You should at least define 1 channel or shmem."
#endif /* (DUALOSCOM_NUM_CHANNELS == 0) && (DUALOSCOM_NUM_SHMEMS == 0) */

// #define DUALOSCOM_DEBUG

#define SHREGION_CONTROL_SIZE  (4*3)

/*
 * INTERNAL TYPES
 */
typedef struct {
    uint32_t *p_write;
    uint32_t *p_read;
    uint32_t *p_buffer;
    uint32_t  num_elements;
} dualoscom_fifo_t;

typedef struct {
    bool_t                   mutex_protection;
    uint32_t                 block_size;
    uint32_t                 num_blocks;
    uint32_t                *p_blocks_buffer;
    dualoscom_fifo_t         t2nt_fifo;
    dualoscom_fifo_t         nt2t_fifo;
    dualoscom_filter_id_t   *p_t2nt_filter_id;
    dualoscom_filter_id_t   *p_nt2t_filter_id;
} dualoscom_channel_t;

typedef struct {
    uint32_t                *p_lock;
    uint32_t                 buffer_size;
    uint32_t                *p_buffer;
    uint32_t                 channel_type;
} dualoscom_shmem_t;

#ifdef DUALOSCOM_T

#define DUALOSCOM_NUM_SRC2DEST_FILTERS DUALOSCOM_NUM_T2NT_FILTERS
#define DUALOSCOM_NUM_DEST2SRC_FILTERS DUALOSCOM_NUM_NT2T_FILTERS

#define GET_SRC2DEST_FIFO(ch) (ch->t2nt_fifo)
#define GET_DEST2SRC_FIFO(ch) (ch->nt2t_fifo)
#define GET_SRC2DEST_FILTER_ID(ch) (ch->p_t2nt_filter_id)
#define GET_DEST2SRC_FILTER_ID(ch) (ch->p_nt2t_filter_id)

#define DUALOSCOM_CHANNEL_DEFAULT_DEST2SRC_FILTERS DUALOSCOM_CHANNEL_DEFAULT_NT2T_FILTERS

#else /* DUALOSCOM_NT */

#define DUALOSCOM_NUM_SRC2DEST_FILTERS DUALOSCOM_NUM_NT2T_FILTERS
#define DUALOSCOM_NUM_DEST2SRC_FILTERS DUALOSCOM_NUM_T2NT_FILTERS

#define GET_SRC2DEST_FIFO(ch) (ch->nt2t_fifo)
#define GET_DEST2SRC_FIFO(ch) (ch->t2nt_fifo)
#define GET_SRC2DEST_FILTER_ID(ch) (ch->p_nt2t_filter_id)
#define GET_DEST2SRC_FILTER_ID(ch) (ch->p_t2nt_filter_id)

#define DUALOSCOM_CHANNEL_DEFAULT_DEST2SRC_FILTERS DUALOSCOM_CHANNEL_DEFAULT_T2NT_FILTERS

#endif /* DUALOSCOM_T */

typedef struct {
    bool_t initialized;
    dualoscom_channel_t channels[DUALOSCOM_NUM_CHANNELS];
    dualoscom_shmem_t   shmems[DUALOSCOM_NUM_SHMEMS];
    dualoscom_filter_t  filters[DUALOSCOM_NUM_SRC2DEST_FILTERS];
} dualoscom_user_t;

/*
 * MODULE VARIABLES
 */
static dualoscom_user_t the_dualoscom_user = {
    .initialized = false,
#ifdef DUALOSCOM_FILTERS
    .filters     = DUALOSCOM_FILTERS,
#endif /* DUALOSCOM_FILTERS */
};


/*
 *  GET control block
 */
#define get_channel(channel_id) (&(the_dualoscom_user.channels[channel_id]))
#define get_shmem(shmem_id) (&(the_dualoscom_user.shmems[shmem_id]))

/*
 * FIFO
 */
static inline int __dualoscom_fifo_empty(dualoscom_fifo_t *fifo)
{
    return (*fifo->p_write == *fifo->p_read);
}

#ifdef COPY_UPDATE_ALGORITHM_ENABLE

static inline int __dualoscom_fifo_full(dualoscom_fifo_t *fifo)
{
    uint32_t fifo_write;
    uint32_t fifo_read;

    fifo_write = *fifo->p_write;
    assert(fifo_write >= 0);
    assert(fifo_write < fifo->num_elements);

    fifo_read = *fifo->p_read;
    assert(fifo_read >= 0);
    assert(fifo_read < fifo->num_elements);

    return ((fifo_write + 1) % fifo->num_elements) == fifo_read;
}

static inline int __dualoscom_fifo_enqueue(dualoscom_fifo_t *fifo, uint32_t value)
{
    uint32_t fifo_write;

    if (__dualoscom_fifo_full(fifo)) return -1;

    fifo_write = *fifo->p_write;
    assert(fifo_write >= 0);
    assert(fifo_write < fifo->num_elements);

    fifo->p_buffer[fifo_write] = value;
    *fifo->p_write = (fifo_write + 1) % fifo->num_elements;

    return 0;
}

static inline int __dualoscom_fifo_dequeue(dualoscom_fifo_t *fifo, uint32_t *value)
{
    uint32_t fifo_read;

    if (__dualoscom_fifo_empty(fifo)) return -1;

    fifo_read = *fifo->p_read;
    assert(fifo_read >= 0);
    assert(fifo_read < fifo->num_elements);

    *value = fifo->p_buffer[*fifo->p_read];
    *fifo->p_read = (*fifo->p_read + 1) % fifo->num_elements;

    return 0;
}

#else /* !COPY_UPDATE_ALGORITHM_ENABLE */

static inline int __dualoscom_fifo_full(dualoscom_fifo_t *fifo)
{
    return ((*fifo->p_write + 1) % fifo->num_elements) == *fifo->p_read;
}

static inline int __dualoscom_fifo_enqueue(dualoscom_fifo_t *fifo, uint32_t value)
{
    if (__dualoscom_fifo_full(fifo)) return -1;

    fifo->p_buffer[*fifo->p_write] = value;
    *fifo->p_write = (*fifo->p_write + 1) % fifo->num_elements;

    return 0;
}

static inline int __dualoscom_fifo_dequeue(dualoscom_fifo_t *fifo, uint32_t *value)
{
    if (__dualoscom_fifo_empty(fifo)) return -1;

    *value = fifo->p_buffer[*fifo->p_read];
    *fifo->p_read = (*fifo->p_read + 1) % fifo->num_elements;

    return 0;
}

#endif /* COPY_UPDATE_ALGORITHM_ENABLE */

/*
 *  CHECK
 */
#define CHECK_NOINIT() { \
    if (the_dualoscom_user.initialized == false) { \
        return DUALOSCOM_NOINIT; \
    } \
}

#define CHECK_CHANNEL_ID(id) { \
    if (id >= DUALOSCOM_NUM_CHANNELS) { \
        return DUALOSCOM_PARAM; \
    } \
}

#define CHECK_SHMEM_ID(id) { \
    if (id >= DUALOSCOM_NUM_SHMEMS) { \
        return DUALOSCOM_PARAM; \
    } \
}

#define CHECK_BLOCK_ID(id, p_ch) { \
    if (id >= p_ch->num_blocks) { \
        return DUALOSCOM_PARAM; \
    } \
}

#define CHECK_NULL(ptr) { \
    if (ptr == NULL) { \
        return DUALOSCOM_PARAM; \
    } \
}

#define CHECK_FILTER_ID(id) { \
    if ((id >= DUALOSCOM_NUM_DEST2SRC_FILTERS) && (id != NULL_FILTER)) { \
        return DUALOSCOM_PARAM; \
    } \
}

#define CHECK_GROUP_ID(id) { \
    if (id >= DUALOSCOM_NUM_GROUPS) { \
       return DUALOSCOM_PARAM; \
    } \
}

/*
 * DEBUG
 */
#ifdef DUALOSCOM_DEBUG

#define DUALOSCOM_DEBUG_INIT
#define DUALOSCOM_DEBUG_ALLOC
#define DUALOSCOM_DEBUG_GETBUFFER
#define DUALOSCOM_DEBUG_ENQUEUE
#define DUALOSCOM_DEBUG_DEQUEUE

static void ___dualoscom_debug_print_locks(dualoscom_channel_t *ch)
{
    int i;
    uint32_t *tmp;

    PRINT("LOCKS");
    for(i=0; i<ch->num_blocks; i++) {
        tmp = ch->p_blocks_buffer + i*(1 + ch->block_size);
        PRINT("*(0x%x): 0x%x", (uint32_t)tmp, *((uint32_t *)tmp));
    }
}

static void ___dualoscom_debug_print_channel(dualoscom_channel_t *ch)
{
    PRINT("mutex_protection: %s", (ch->mutex_protection) ? "true" : "false");
    PRINT("block_size: %u words", ch->block_size);
    PRINT("num_blocks: %u", ch->num_blocks);
    PRINT("p_blocks_buffer: 0x%x", (uint32_t)ch->p_blocks_buffer);
    PRINT("p_t2nt_filter_id: %u", *ch->p_t2nt_filter_id);
    PRINT("p_nt2t_filter_id: %u", *ch->p_nt2t_filter_id);
}

static void ___dualoscom_debug_print_fifos(dualoscom_channel_t *ch)
{
    int i;

    PRINT("t2nt_fifo.p_write: %u", *ch->t2nt_fifo.p_write);
    PRINT("t2nt_fifo.p_read: %u", *ch->t2nt_fifo.p_read);
    PRINT("t2nt_fifo.p_buffer: 0x%x", (uint32_t)ch->t2nt_fifo.p_buffer);
    PRINT("t2nt_fifo.num_elements: %u", (uint32_t)ch->t2nt_fifo.num_elements);
    PRINT("t2nt_fifo:");
    for(i=0; i<ch->num_blocks+1; i++) {
        PRINT("*(0x%x): 0x%x", ch->t2nt_fifo.p_buffer + i, *(ch->t2nt_fifo.p_buffer + i));
    }

    PRINT("nt2t_fifo.p_write: %u", *ch->nt2t_fifo.p_write);
    PRINT("nt2t_fifo.p_read: %u", *ch->nt2t_fifo.p_read);
    PRINT("nt2t_fifo.p_buffer: 0x%x", (uint32_t)ch->nt2t_fifo.p_buffer);
    PRINT("nt2t_fifo.num_elements: %u", (uint32_t)ch->nt2t_fifo.num_elements);
    PRINT("nt2t_fifo:");
    for(i=0; i<ch->num_blocks+1; i++) {
        PRINT("*(0x%x): 0x%x", ch->nt2t_fifo.p_buffer + i, *(ch->nt2t_fifo.p_buffer + i));
    }
}

static void ___dualoscom_debug_print_blocks(dualoscom_channel_t *ch)
{
    int i, j;
    uint32_t *tmp, *tmp2;

    PRINT("BLOCKS");
    for(i=0; i<ch->num_blocks; i++) {
        tmp = ch->p_blocks_buffer + i*(1 + ch->block_size) + 1;
        PRINT("block(%d): 0x%x:", i, (uint32_t)tmp);
        for(j=0; j<ch->block_size; j++) {
            tmp2 = tmp + j;
            PRINT("*(0x%X): 0x%x", (uint32_t)(tmp2), *(tmp2));
        }
    }
}

static void ___dualoscom_debug_print_all()
{
    int i;
    dualoscom_channel_t *ch;

    PRINT("initialized: %s", (the_dualoscom_user.initialized)?"true":"false");

    for(i=0; i<DUALOSCOM_NUM_SRC2DEST_FILTERS; i++)
        PRINT("filter(%d): 0x%x", i, (uint32_t)the_dualoscom_user.filters[i]);

    for(i=0; i<DUALOSCOM_NUM_CHANNELS; i++) {
        ch = &the_dualoscom_user.channels[i];
        PRINT("CHANNEL %d", i);
        ___dualoscom_debug_print_channel(ch);
        ___dualoscom_debug_print_fifos(ch);
        ___dualoscom_debug_print_locks(ch);
        ___dualoscom_debug_print_blocks(ch);
    }
}
#endif /* DUALOSCOM_DEBUG */

#ifdef DUALOSCOM_DEBUG_ALLOC
#define __dualoscom_debug_print_locks(ch) ___dualoscom_debug_print_locks(ch) 
#else /* !DUALOSCOM_DEBUG_ALLOC */
#define __dualoscom_debug_print_locks(ch)
#endif /* DUALOSCOM_DEBUG_ALLOC */

#ifdef DUALOSCOM_DEBUG_GETBUFFER
#define __dualoscom_debug_print_channel(ch)  ___dualoscom_debug_print_channel(ch)
#else /* !DUALOSCOM_DEBUG_GETBUFFER */
#define __dualoscom_debug_print_channel(ch)
#endif /* DUALOSCOM_DEBUG_GETBUFFER */

#ifdef DUALOSCOM_DEBUG_ENQUEUE
#define __dualoscom_debug_print_fifos(ch)   ___dualoscom_debug_print_fifos(ch)
#else  /* !DUALOSCOM_DEBUG_ENQUEUE */
#define __dualoscom_debug_print_fifos(ch)
#endif /* DUALOSCOM_DEBUG_ENQUEUE */

#ifdef DUALOSCOM_DEBUG_INIT
#define __dualoscom_debug_print_all() ___dualoscom_debug_print_all()
#else  /* !DUALOSCOM_DEBUG_INIT */
#define __dualoscom_debug_print_all()
#endif /* DUALOSCOM_DEBUG_INIT */

/*
 * dualoscom_init
 *
 * Initializates the Dual-OS Communications system.
 *
 * Errors: DUALOSCOM_PARAM, DUALOSCOM_TIMEOUT, DUALOSCOM_ESHREGION, DUALOSCOM_ECTX
 *
 */
int dualoscom_init(bool_t master, dualoscom_time_t timeout)
{
    int ret, i;
    void *p_shregion;
    dualoscom_channel_t *p_ch;
    dualoscom_shmem_t *p_shmem;

    uint32_t num_blocks[DUALOSCOM_NUM_CHANNELS] = DUALOSCOM_CHANNEL_NUM_BLOCKS;
    uint32_t block_size[DUALOSCOM_NUM_CHANNELS] = DUALOSCOM_CHANNEL_BLOCK_SIZES;
    uint32_t def_filter[DUALOSCOM_NUM_CHANNELS] = DUALOSCOM_CHANNEL_DEFAULT_DEST2SRC_FILTERS;
    uint32_t channel_offsets[DUALOSCOM_NUM_CHANNELS] = DUALOSCOM_SHREGION_CHANNEL_OFFSETS;

    uint32_t buffer_size[DUALOSCOM_NUM_SHMEMS] = DUALOSCOM_SHMEM_BUFFER_SIZES;
    uint32_t shmem_offsets[DUALOSCOM_NUM_SHMEMS]  = DUALOSCOM_SHREGION_SHMEM_OFFSETS;

    if (!__dualoscom_kernel_check_tskctx()) return DUALOSCOM_ECTX;
    
    if (!master) return 0; // TODO: complete this

    if(the_dualoscom_user.initialized == true)  return DUALOSCOM_SUCCESS;

    // 1.- create shared memory object and map it and 0 clean.
    if (__dualoscom_kernel_shregion_init(&p_shregion) != 0) return DUALOSCOM_ESHREGION;
    
    // 2.- initialize all channels
    for(i=0; i<DUALOSCOM_NUM_CHANNELS; i++) {
        p_ch = get_channel(i);
        // local variables
        p_ch->block_size = block_size[i];
        p_ch->num_blocks = num_blocks[i];
        p_ch->t2nt_fifo.num_elements = p_ch->num_blocks + 1;
        p_ch->nt2t_fifo.num_elements = p_ch->num_blocks + 1;
        // initialize pointers to shared memory
        p_ch->t2nt_fifo.p_write  = p_shregion + channel_offsets[i];
        p_ch->t2nt_fifo.p_read   = p_ch->t2nt_fifo.p_write  + 1;
        p_ch->t2nt_fifo.p_buffer = p_ch->t2nt_fifo.p_read   + 1;
        p_ch->nt2t_fifo.p_write  = p_ch->t2nt_fifo.p_buffer +  p_ch->t2nt_fifo.num_elements;
        p_ch->nt2t_fifo.p_read   = p_ch->nt2t_fifo.p_write  + 1;
        p_ch->nt2t_fifo.p_buffer = p_ch->nt2t_fifo.p_read   + 1;
        p_ch->p_t2nt_filter_id   = p_ch->nt2t_fifo.p_buffer + p_ch->nt2t_fifo.num_elements;
        p_ch->p_nt2t_filter_id   = p_ch->p_t2nt_filter_id   + 1;
        p_ch->p_blocks_buffer    = p_ch->p_nt2t_filter_id   + 1;

        // set default filter
        *GET_DEST2SRC_FILTER_ID(p_ch) = def_filter[i];
        // Note: pending events are already 0 due to the memset
    }

    // 2.- initialize all shmems
    for(i=0; i<DUALOSCOM_NUM_SHMEMS; i++) {
        p_shmem = get_shmem(i);
        p_shmem->buffer_size = buffer_size[i];
        p_shmem->p_lock = p_shregion + shmem_offsets[i];
        p_shmem->p_buffer = p_shmem->p_lock + 1;
    }

    __dualoscom_debug_print_all();

    // 4.- init events and wait for the Destinataion OS to initialize (with a timeout)
    ret = __dualoscom_kernel_init(timeout, p_shregion);
    if (ret != 0) return DUALOSCOM_TIMEOUT;

    the_dualoscom_user.initialized = true;

    return DUALOSCOM_SUCCESS;
}

/*
 * dualoscom_shmem_get
 *
 * To obtain a pointer to the beginning of the memory region of a shmem.
 * 
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM
 *
 */
int dualoscom_shmem_get(const dualoscom_shmem_id_t shmem_id,
                               void **buffer, uint32_t *size)
{
    dualoscom_shmem_t *p_shmem;

    CHECK_NOINIT();
    CHECK_SHMEM_ID(shmem_id);
    p_shmem = get_shmem(shmem_id);

    *buffer = (void *)(p_shmem->p_buffer);
    *size   = p_shmem->buffer_size;

    return DUALOSCOM_SUCCESS;
}

/*
 * dualoscom_shmem_trylock
 *
 * Try to get lock of a shmem.
 * 
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM, DUALOSCOM_LOCKED
 *
 */
int dualoscom_shmem_trylock(const dualoscom_shmem_id_t shmem_id)
{
    dualoscom_shmem_t *p_shmem;
    uint32_t *p_lock;
    int ret;

    CHECK_NOINIT();
    CHECK_SHMEM_ID(shmem_id);
    p_shmem = get_shmem(shmem_id);

    p_lock = p_shmem->p_lock;
    for(;;){
        if (*p_lock == 1) return DUALOSCOM_LOCKED;
        ret = __sync_bool_compare_and_swap (p_lock, 0, 1);
        if (ret == true) break;
    }

    return DUALOSCOM_SUCCESS;
}

/*
 * dualoscom_shmem_unlock
 *
 * Release lock of a shmem.
 * 
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM, DUALOSCOM_UNLOCKED
 *
 */
int dualoscom_shmem_unlock(const dualoscom_shmem_id_t shmem_id)
{
    dualoscom_shmem_t *p_shmem;
    uint32_t *p_lock;

    CHECK_NOINIT();
    CHECK_SHMEM_ID(shmem_id);
    p_shmem = get_shmem(shmem_id);

    p_lock = p_shmem->p_lock;
    if (*p_lock == 0) return DUALOSCOM_UNLOCKED;
    *p_lock = 0;

    return DUALOSCOM_SUCCESS;
}

/*
 * dualoscom_channel_block_alloc
 *
 * Allocates a block from a channel's pool. This function never blocks the
 * calling task or process.
 *
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM, DUALOSCOM_FULL
 *
 */
int dualoscom_channel_block_alloc(const dualoscom_channel_id_t   channel_id,
                                  dualoscom_block_id_t          *block_id)
{
    int i, ret;
    dualoscom_channel_t *p_ch;
    uint32_t *p_lock;

    CHECK_NOINIT();
    CHECK_CHANNEL_ID(channel_id);
    CHECK_NULL(block_id);
    p_ch = get_channel(channel_id);

    for(i=0; i < p_ch->num_blocks; i++) {
        p_lock = p_ch->p_blocks_buffer + i*(1 + p_ch->block_size);
        // atomic: if (*p_lock == 0) *p_lock = 1, ret = true
        ret = __sync_bool_compare_and_swap (p_lock, 0, 1);
        if (ret == false) continue;
        *block_id = (dualoscom_block_id_t)i;

        __dualoscom_debug_print_locks(p_ch);

        return DUALOSCOM_SUCCESS;
    }

    __dualoscom_debug_print_locks(p_ch);

    return DUALOSCOM_FULL;
}

/*
 * dualoscom_channel_block_free
 *
 * Releases a block back to the channel's pool where it belongs.
 *
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM, DUALOSCOM_ALLOC
 *
 */
int dualoscom_channel_block_free(const dualoscom_channel_id_t channel_id,
                                 const dualoscom_block_id_t block_id)
{
    dualoscom_channel_t *p_ch;
    uint32_t *p_lock;

    CHECK_NOINIT();
    CHECK_CHANNEL_ID(channel_id);
    p_ch = get_channel(channel_id);
    CHECK_BLOCK_ID(block_id, p_ch)

    p_lock = p_ch->p_blocks_buffer + block_id*(1 + p_ch->block_size);
    if (*p_lock == 0) return DUALOSCOM_ALLOC;

    *p_lock = 0;

    __dualoscom_debug_print_locks(p_ch);

    return DUALOSCOM_SUCCESS;
}

/*
 * dualoscom_channel_block_getbuffer
 *
 * To obtain a pointer to the beginning of the memory region of a block.
 *
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM, DUALOSCOM_ALLOC
 *
 */
int dualoscom_channel_block_getbuffer(const dualoscom_channel_id_t channel_id,
                                      const dualoscom_block_id_t   block_id,
                                      void                       **buffer,
                                      uint32_t                    *size)
{
    dualoscom_channel_t *p_ch;
    uint32_t *p_lock;

    CHECK_NOINIT();
    CHECK_CHANNEL_ID(channel_id);
    p_ch = get_channel(channel_id);
    CHECK_BLOCK_ID(block_id, p_ch)

    p_lock = p_ch->p_blocks_buffer + block_id*(1 + p_ch->block_size);
    if (*p_lock == 0) return DUALOSCOM_ALLOC;

    *buffer = (void *)(p_lock + 1);
    *size   = p_ch->block_size;

    __dualoscom_debug_print_channel(p_ch);

    return DUALOSCOM_SUCCESS;
}

/*
 * dualoscom_channel_block_enqueue
 *
 * Enqueues a block to a channel's FIFO. Note that the channel is implicit in
 * the block identifier.
 *
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM, DUALOSCOM_FILTER, DUALOSCOM_ALLOC.
 *
 */
int dualoscom_channel_block_enqueue(const dualoscom_channel_id_t channel_id,
                                    const dualoscom_block_id_t block_id)
{
    dualoscom_channel_t *p_ch;
    uint32_t *p_lock;
    int ret;
    dualoscom_filter_id_t filter_id;
    dualoscom_filter_t filter;
    void *buffer;

    CHECK_NOINIT();
    CHECK_CHANNEL_ID(channel_id);
    p_ch = get_channel(channel_id);
    CHECK_BLOCK_ID(block_id, p_ch)

    // TODO: mutex

    // filtering
    filter_id = *GET_SRC2DEST_FILTER_ID(p_ch);
    if (filter_id != NULL_FILTER ) {
        assert(filter_id < DUALOSCOM_NUM_SRC2DEST_FILTERS);
        filter = the_dualoscom_user.filters[filter_id];
        // TODO: check this address
        buffer = p_ch->p_blocks_buffer + block_id*(1 + p_ch->block_size) + 1;
        if (filter(buffer, p_ch->block_size) == false) return DUALOSCOM_FILTER;
    }
    p_lock = p_ch->p_blocks_buffer + block_id*(1 + p_ch->block_size);
    if (*p_lock == 0) return DUALOSCOM_ALLOC;

    __dualoscom_kernel_lock();
    ret = __dualoscom_fifo_enqueue(&GET_SRC2DEST_FIFO(p_ch), block_id);
    __dualoscom_kernel_unlock();
    assert(ret == 0);

    __dualoscom_debug_print_fifos(p_ch);

    return 0;
}

/*
 * dualoscom_channel_block_dequeue
 *
 * Dequeues a block from a channel's FIFO. This function never blocks the
 * calling task or process.
 *
 * Errors: DUALOSCOM_NOPERM, DUALOSCOM_NOINIT, DUALOSCOM_PARAM,
 *         DUALOSCOM_EMPTY, DUALOSCOM_ALLOC
 *
 */
int dualoscom_channel_block_dequeue(const dualoscom_channel_id_t   channel_id,
                                    dualoscom_block_id_t          *block_id)
{
    dualoscom_channel_t *p_ch;
    int ret;
    uint32_t *p_lock;

    // TODO: mutex
    CHECK_NOINIT();
    CHECK_CHANNEL_ID(channel_id);
    CHECK_NULL(block_id);
    p_ch = get_channel(channel_id);

    __dualoscom_kernel_lock();
    ret = __dualoscom_fifo_dequeue(&GET_DEST2SRC_FIFO(p_ch), block_id);
    __dualoscom_kernel_unlock();
    if (ret == -1) return DUALOSCOM_EMPTY;

    assert(*block_id < p_ch->num_blocks);

    p_lock = p_ch->p_blocks_buffer + *block_id*(1 + p_ch->block_size);
    if (*p_lock == 0) return DUALOSCOM_ALLOC;

    __dualoscom_debug_print_fifos(p_ch);

    return 0;
}

/*
 * dualoscom_channel_event_send
 *
 * Sends an asynchronous channel event notification. If a notification
 * was already sent and has not been acknowledged yet by the receiver, the
 * function will return as if the notification was sent successfully.
 *
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM
 *
 */
int dualoscom_channel_event_send(const dualoscom_channel_id_t channel_id)
{
    int ret;

    CHECK_NOINIT();
    CHECK_CHANNEL_ID(channel_id);

    ret = __dualoscom_kernel_sendevent(channel_id);
    assert(ret == 0);

    return 0;
}

/*
 * dualoscom_channel_event_wait
 *
 * This function makes the calling task - or process - wait for an
 * asynchronous event notification on a specific channel. If an event was
 * pending, the function acknowledges it and returns immediately.
 * Otherwise, the calling thread is put in waiting state until an event
 * arrives or a timeout occurs. The units in which the timeout is
 * expressed are implementation-dependent.
 *
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM, DUALOSCOM_TIMEOUT, DUALOSCOM_RESTART
 *
 */
int dualoscom_channel_event_wait(const dualoscom_channel_id_t channel_id,
                                 const dualoscom_time_t       timeout)
{
    int ret;

    CHECK_NOINIT();
    CHECK_CHANNEL_ID(channel_id);

    if (!__dualoscom_kernel_check_tskctx()) return DUALOSCOM_ECTX;
    
    ret = __dualoscom_kernel_waitevent(channel_id, timeout);

    return ret;
}

/*
 * dualoscom_channel_filter_set
 *
 * Used by the receiver to set a communications filter.
 *
 * Errors: DUALOSCOM_NOINIT, DUALOSCOM_PARAM
 *
 */
int dualoscom_channel_filter_set(const dualoscom_channel_id_t channel_id,
                                  const dualoscom_filter_id_t  filter_id)
{
    dualoscom_channel_t *p_ch;

    CHECK_NOINIT();
    CHECK_CHANNEL_ID(channel_id);
    CHECK_FILTER_ID(filter_id);
    p_ch = get_channel(channel_id);

    *GET_DEST2SRC_FILTER_ID(p_ch) = filter_id;

    return DUALOSCOM_SUCCESS;
}

/*
 * dualoscom_channel_group_event_wait  (under development)
 *
 * This function makes the calling task - or process - wait for an
 * asynchronous event notification on a specific group. If an event was
 * pending, the function acknowledges it and returns immediately.
 * Otherwise, the calling thread is put in waiting state until an event
 * arrives or a timeout occurs. The units in which the timeout is
 * expressed are implementation-dependent.
 *
 * Errors: DUALOSCOM_NOPERM, DUALOSCOM_NOINIT, DUALOSCOM_PARAM,
 *         DUALOSCOM_TIMEOUT
 *
 */
int dualoscom_group_event_wait(const dualoscom_group_id_t group_id,
                               const dualoscom_time_t     timeout)
{
    int ret;

    CHECK_NOINIT();
#ifdef DUALOSCOM_NUM_GROUPS
    CHECK_GROUP_ID(group_id);
#endif /* DUALOSCOM_NUM_GROUPS */

    ret = __dualoscom_kernel_waitgroup(group_id, timeout);
    assert(ret == 0);

    return 0;
}


/*
 * Initialization for the dualoscom system retart.
 */
void
dualoscom_restart(void) {
    the_dualoscom_user.initialized = false;
}
