#include "dualoscom.h"
#include "dualoscom_kernel.h"

/*
 * KERNEL INTERFACE
 */
static int fd;
pthread_mutex_t mutex; 

/*
 * INIT Shared Region(Memory)
 */
int __dualoscom_kernel_shregion_init(void **shregion)
{
    fd = open("/dev/safeg", O_RDWR);
    assert(fd >= 0);

    *shregion = mmap(NULL, DUALOSCOM_SHREGION_SIZE, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
    assert(*shregion != MAP_FAILED);
    // Note: the shared memory is zero-ed by the RTOS at initialization

    return 0;
}

int __dualoscom_kernel_init(dualoscom_time_t timeout, void *shregion)
{
    dualoscom_init_data_t init_data;
#ifdef DUALOSCOM_NUM_GROUPS
    bool channels2groups[DUALOSCOM_NUM_CHANNELS][DUALOSCOM_NUM_GROUPS] = DUALOSCOM_CHANNEL2GROUPS;
    init_data.num_groups           = DUALOSCOM_NUM_GROUPS;
    init_data.channels2groups      = channels2groups;
#else /* !DUALOSCOM_NUM_GROUPS */
    init_data.num_groups           = 0;
    init_data.channels2groups      = NULL;
#endif /* DUALOSCOM_NUM_GROUPS */
    init_data.num_channels         = DUALOSCOM_NUM_CHANNELS;
    init_data.num_shmem            = DUALOSCOM_NUM_SHMEMS;
    init_data.shmem_pending_events = DUALOSCOM_SHREGION_PENDING_EVENTS;
#if 0
    DEBUG(DUALOSCOM_DEBUG_INIT, "num_channels:%u, num_groups:%u, \
            shmem_pending:0x%X channels2groups:0x%X\n",
            init_data.num_channels, init_data.num_groups,
            init_data.shmem_pending_events,
            (uint32_t)init_data.channels2groups);
#endif
    pthread_mutex_init(&mutex, NULL);
    return ioctl(fd, DUALOSCOM_IOCTL_CMD_INIT, &init_data);
}

int __dualoscom_kernel_sendevent(dualoscom_channel_id_t channel_id)
{
    return ioctl(fd, DUALOSCOM_IOCTL_CMD_SENDEVENT, channel_id);
}

int __dualoscom_kernel_waitevent(dualoscom_channel_id_t channel_id,
                                        dualoscom_time_t timeout)
{
    return ioctl(fd, DUALOSCOM_IOCTL_CMD_WAITEVENT, channel_id);
}

int __dualoscom_kernel_waitgroup(dualoscom_group_id_t group_id, dualoscom_time_t timeout){ return -1;};
