#ifndef DUALOSCOM_KERNEL_H
#define DUALOSCOM_KERNEL_H

#include <unistd.h>    /* ftruncate */
#include <fcntl.h>     /* flags: O_CREAT, etc */
#include <sys/stat.h>  /* modes: S_IRWXU, etc */
#include <sys/mman.h>  /* shm_xxx and mmap() */
#include <semaphore.h> /* sem_xxx */
#include <assert.h>    /* assert() */
#include <string.h>    /* memset */
#include <pthread.h>   /* pthread_xxx */
#include <sys/ioctl.h>
#include <stdio.h>
#include "nt_os-dualoscom_config.h"

#define DUALOSCOM_NT

#define DEBUG(enable,x,args...) \
        if(enable) printf("DEBUG (%s): " x, __func__ , ##args)

#define PRINT(x, args...) \
    printf(x, ##args);

/*
 * KERNEL INTERFACE
 */
extern int __dualoscom_kernel_init(dualoscom_time_t timeout, void *shregion);
extern int __dualoscom_kernel_shregion_init(void **shregion);
extern int __dualoscom_kernel_sendevent(dualoscom_channel_id_t channel_id);
extern int __dualoscom_kernel_waitevent(dualoscom_channel_id_t channel_id, dualoscom_time_t timeout);
extern int __dualoscom_kernel_waitgroup(dualoscom_group_id_t group_id, dualoscom_time_t timeout);

static inline bool_t
__dualoscom_kernel_check_tskctx()
{
    return true;
}

extern pthread_mutex_t mutex;

static inline void
__dualoscom_kernel_lock()
{
    pthread_mutex_lock(&mutex);
}

static inline void
__dualoscom_kernel_unlock()
{
    pthread_mutex_unlock(&mutex);
}


#if 0
static inline int __sync_bool_compare_and_swap(uint32_t *a, uint32_t old, uint32_t future)
{
    uint32_t tmp, loaded;

    asm volatile(
        "1:                                   \n\t"
        "ldrex   %[loaded], [%[atomic]]       \n\t"
        "cmp     %[loaded], %[old]            \n\t"
        "bne     2f                           \n\t"
        "strex   %[tmp], %[future], [%[atomic]]  \n\t"
        "tst     %[tmp], #1                   \n\t"
        "bne     1b                           \n\t"
        "2:                                   \n\t"
        : [tmp] "=&r" (tmp), [loaded] "=&r" (loaded), "=m" (*a)
        : [old] "r" (old), [future] "r" (future), [atomic] "r" (a)
        );

    return loaded == old;
}
#endif
#endif /* DUALOSCOM_KERNEL_H */
