/*
 *  TOPPERS/SafeG Dual-OS monitor
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Safety Gate Dual-OS monitor
 *
 *  Copyright (C) 2012-2013 by Embedded and Real-Time Systems Laboratory
 *     Graduate School of Information Science, Nagoya University, JAPAN
 *
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 *
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 *
 *  @(#) $Id: dualoscom_kernel.c 428 2013-03-05 08:38:58Z ertl-honda $
 */

#include "dualoscom.h"
#include "dualoscom_kernel.h"
#include "kernel_cfg.h"

/*
 * CONTROL OFFSET
 */
#define CTL_T2NT_REQ  0x00
#define CTL_NT2T_REQ  0x04

#define CTL_NT2T_REQ_APPEXIT 0x01
#define CTL_NT2T_REQ_EVENT   0x02
#define CTL_NT2T_INIT        0x87654321

#define CTL_T2NT_INIT        0x12345678


/*
 *  initialized task id;
 */
static ID init_task_id;

/*
 * INTERNAL TYPES
 */
typedef struct {
    ID       rtos_sem;
    ID       rtos_tskid;
    uint32_t *t2nt_pending;
    uint32_t *nt2t_pending;
#ifdef DUALOSCOM_NUM_GROUPS
    bool_t   groups[DUALOSCOM_NUM_GROUPS];
#endif /* DUALOSCOM_NUM_GROUPS */
} dualoscom_channel_event_t;

typedef struct {
    ID        rtos_sem;
    uint32_t  waiting;
} dualoscom_group_event_t;

typedef struct {
    bool_t initialized;
    dualoscom_channel_event_t channel_events[DUALOSCOM_NUM_CHANNELS];
#ifdef DUALOSCOM_NUM_GROUPS
    dualoscom_group_event_t group_events[DUALOSCOM_NUM_GROUPS];
#endif /* DUALOSCOM_NUM_GROUPS */
} dualoscom_kernel_t;

/*
 * DEBUG
 */
// #define DUALOSCOM_DEBUG

#define DEBUG_OUT false

#ifdef DUALOSCOM_DEBUG

static void __dualoscom_debug_print_chevent(dualoscom_channel_event_t *ev)
{
    ER      ercd;
    T_RSEM  rsem;

    syslog(LOG_NOTICE, "chevent->t2nt_pending: 0x%X (val:0x%X)\n", (uint32_t)ev->t2nt_pending, *(ev->t2nt_pending));
    syslog(LOG_NOTICE, "chevent->nt2t_pending: 0x%X (val:0x%X)\n", (uint32_t)ev->nt2t_pending, *(ev->nt2t_pending));

    ercd = ref_sem(ev->rtos_sem, &rsem);
    assert(ercd == E_OK);

    if (rsem.wtskid == TSK_NONE) {
        syslog(LOG_NOTICE, "rtos_sem no task waiting");
    } else {
        syslog(LOG_NOTICE, "rtos_sem task with id %d waiting", rsem.wtskid);
    }
    syslog(LOG_NOTICE, "rtos_sem value: %d", rsem.semcnt);
}

#else /* !DUALOSCOM_DEBUG */
#define  __dualoscom_debug_print_chevent(ev)
#endif /* DUALOSCOM_DEBUG */

/*
 * MODULE VARIABLES
 */
static dualoscom_kernel_t the_dualoscom_kernel;

/*
 * EVENTS HANDLER
 */
void dualoscom_handler(void)
{
    ER ercd;
    int i;
    dualoscom_channel_event_t *chevent;
#ifdef DUALOSCOM_NUM_GROUPS
    int j;
    dualoscom_group_event_t *grevent;
#endif /* DUALOSCOM_NUM_GROUPS */
    bool_t group_signaled;

    DEBUG(DEBUG_OUT, "Dualoscom handler");

    //Initialize
    if ((the_dualoscom_kernel.initialized == false) &&
        (*(uint32_t *)(SHREGION_PHYS_ADDRESS + CTL_NT2T_REQ) == CTL_NT2T_INIT)) {
        *(uint32_t *)(SHREGION_PHYS_ADDRESS + CTL_T2NT_REQ) = 0;
        *(uint32_t *)(SHREGION_PHYS_ADDRESS + CTL_NT2T_REQ) = 0;
        ercd = isig_sem(DUALOSCOM_SEM_INIT);
        assert(ercd == E_OK);
        the_dualoscom_kernel.initialized = true;
        return;
    }
    //NT_OS application has been exit. Restart dualos communications system
    else if(*(uint32_t *)(SHREGION_PHYS_ADDRESS + CTL_NT2T_REQ) == CTL_NT2T_REQ_APPEXIT) {
        extern void dualoscom_restart(void);
        dualoscom_restart();
        for(i=0; i<DUALOSCOM_NUM_CHANNELS; i++) {
            chevent = &the_dualoscom_kernel.channel_events[i];
            *(chevent->nt2t_pending) = 0;
            *(chevent->t2nt_pending) = 0;
            if (chevent->rtos_tskid != 0){
                irel_wai(chevent->rtos_tskid);
            }
        }
        *(uint32_t *)(SHREGION_PHYS_ADDRESS + CTL_NT2T_REQ) = 0;
        the_dualoscom_kernel.initialized = false;
        iras_tex(init_task_id, 1);
        return;
    }
    // signal all channels with pending events
    else if(*(uint32_t *)(SHREGION_PHYS_ADDRESS + CTL_NT2T_REQ) == CTL_NT2T_REQ_EVENT) {
        *(uint32_t *)(SHREGION_PHYS_ADDRESS + CTL_NT2T_REQ) = 0;
        for(i=0; i<DUALOSCOM_NUM_CHANNELS; i++) {
            chevent = &the_dualoscom_kernel.channel_events[i];
            if (*(chevent->nt2t_pending) == 0) continue;
            
            // signal group (if a task is waiting)
            group_signaled = false;
#ifdef DUALOSCOM_NUM_GROUPS
            for(j=0; j<DUALOSCOM_NUM_GROUPS; j++) {
                if (chevent->groups[j] == false) continue;
                grevent = &the_dualoscom_kernel.group_events[j];
                if (grevent->waiting == 0) continue;
                isig_sem(grevent->rtos_sem);
                *(chevent->nt2t_pending) = 0;
                group_signaled = true;
            }
#endif /* DUALOSCOM_NUM_GROUPS */
            // otherwise signal the channel
            if (group_signaled == true) continue;
            isig_sem(chevent->rtos_sem);
            *(chevent->nt2t_pending) = 0;
        }
    }
    else {
        assert(0);
    }
}

static void *memset (void *str, int c, size_t len )
{
    register char *st = (char *)str;
    while (len-- > 0) *st++ = c;
    return (void *)str;
}

/*
 * INIT Shared Region(Memory)
 */
int __dualoscom_kernel_shregion_init(void **shregion)
{
    *shregion = (void *)SHREGION_PHYS_ADDRESS;
    memset(*shregion, 0, DUALOSCOM_SHREGION_SIZE);
    return 0;
}

/*
 * INIT
 */
int __dualoscom_kernel_init(dualoscom_time_t timeout, void *shregion)
{
    ER ret;
    int i;
    dualoscom_channel_event_t *chevent;
#ifdef DUALOSCOM_NUM_GROUPS
    int j;
    dualoscom_group_event_t *grevent;
#endif /* DUALOSCOM_NUM_GROUPS */
#ifdef DUALOSCOM_NUM_GROUPS
    bool_t groups[DUALOSCOM_NUM_CHANNELS][DUALOSCOM_NUM_GROUPS] = DUALOSCOM_CHANNEL2GROUPS;
    ID group_semaphores[DUALOSCOM_NUM_GROUPS] = DUALOSCOM_GROUP_SEMAPHORES;
#endif /* DUALOSCOM_NUM_GROUPS */
    ID semaphores[DUALOSCOM_NUM_CHANNELS] = DUALOSCOM_CHANNEL_SEMAPHORES;

    get_tid(&init_task_id);
    
    // 1.- initialize channel events
    for(i=0; i<DUALOSCOM_NUM_CHANNELS; i++) {
        chevent = &the_dualoscom_kernel.channel_events[i];
        chevent->rtos_sem = semaphores[i];
        chevent->t2nt_pending = (uint32_t *)(shregion + DUALOSCOM_SHREGION_PENDING_EVENTS) + i;
        chevent->nt2t_pending = chevent->t2nt_pending + DUALOSCOM_NUM_CHANNELS;

        __dualoscom_debug_print_chevent(chevent);

#ifdef DUALOSCOM_NUM_GROUPS
        for(j=0; j<DUALOSCOM_NUM_GROUPS; j++) {
            chevent->groups[j] = groups[i][j];
        }
#endif /* DUALOSCOM_NUM_GROUPS */
    }

#ifdef DUALOSCOM_NUM_GROUPS
    // 2.- initialize group events
    for(i=0; i<DUALOSCOM_NUM_GROUPS; i++) {
        grevent = &the_dualoscom_kernel.group_events[i];
        grevent->rtos_sem = group_semaphores[i];
        grevent->waiting  = 0;
    }
#endif /* DUALOSCOM_NUM_GROUPS */

    // 3.- wait for the gpos
    the_dualoscom_kernel.initialized = false;
    *(uint32_t *)(SHREGION_PHYS_ADDRESS + CTL_T2NT_REQ) = CTL_T2NT_INIT;
    
    // Flush D-Cache
    // GPOS(or bootlader) will invalidates the D-Cache.
    // Without flushing D-Cache, the data that has been written to the shared memory
    // will not be write back to the shared memory.
    __dualoscom_dcache_flush();

    ret = ena_int(NT_INTNO_SOFTIRQ);
    assert(ret == E_OK);

    ret = twai_sem(DUALOSCOM_SEM_INIT, timeout);
    if (ret == E_TMOUT) return -1;
    
    return 0;
}

int __dualoscom_kernel_sendevent(dualoscom_channel_id_t channel_id)
{
    dualoscom_channel_event_t  *ev;

    ev = &the_dualoscom_kernel.channel_events[channel_id];
    *(ev->t2nt_pending) = 1;

    __dualoscom_send_irq();

    return 0;
}

int __dualoscom_kernel_waitevent(dualoscom_channel_id_t channel_id,
                                 dualoscom_time_t timeout)
{
    dualoscom_channel_event_t  *ev;
    ER ercd;
    ID tskid;

    ev = &the_dualoscom_kernel.channel_events[channel_id];

    __dualoscom_debug_print_chevent(ev);

    if (ev->rtos_tskid != 0) return DUALOSCOM_WAITOVER;
        
    get_tid(&tskid);
    ev->rtos_tskid = tskid;
    ercd = twai_sem(ev->rtos_sem, (TMO)timeout);
    ev->rtos_tskid = 0;

    if (ercd == E_TMOUT) return DUALOSCOM_TIMEOUT;
    if (ercd == E_RLWAI) return DUALOSCOM_RESTART;

    return 0;
}

int __dualoscom_kernel_waitgroup(dualoscom_group_id_t group_id,
                                 dualoscom_time_t timeout)
{
#ifdef DUALOSCOM_NUM_GROUPS
    dualoscom_group_event_t  *gr;

    gr = &the_dualoscom_kernel.group_events[group_id];

    gr->waiting++;
    wai_sem(gr->rtos_sem);
    gr->waiting--;
#endif /* DUALOSCOM_NUM_GROUPS */
    return 0;
}
